import numpy as np

def identity():
    return np.eye(4, dtype=np.float32)

def uniform_scale(s: float):
    return np.matrix([[s, 0, 0, 0], [0, s, 0, 0], [0, 0, s, 0], [0, 0, 0, 1]], dtype=np.float32)

def nonuniform_scale(sx: float, sy: float, sz: float):
    return np.matrix([[sx, 0, 0, 0], [0, sy, 0, 0], [0, 0, sz, 0], [0, 0, 0, 1]], dtype=np.float32)

def translate(tx: float, ty: float, tz: float):
    return np.matrix([[1, 0, 0, tx], [0, 1, 0, ty], [0, 0, 1, tz], [0, 0, 0, 1]], dtype=np.float32)

def rotate_angle_axis(theta: float, v: [float]):
    vx = v[0]
    vy = v[1]
    vz = v[2]
    vxvx = vx * vx
    vxvy = vx * vy
    vxvz = vx * vz
    vyvy = vy * vy
    vyvz = vy * vz
    vzvz = vz * vz
    cosTheta = np.cos(theta)
    sinTheta = np.sin(theta)
    return np.matrix([
        [vxvx * (1 - cosTheta) + cosTheta, vxvy * (1 - cosTheta) - vz * sinTheta, vxvz * (1 - cosTheta) + vy * sinTheta, 0],
        [vxvy * (1 - cosTheta) + vz * sinTheta, vyvy * (1 - cosTheta) + cosTheta, vyvz * (1 - cosTheta) - vx * sinTheta, 0],
        [vxvz * (1 - cosTheta) - vy * sinTheta, vyvz * (1 - cosTheta) + vx * sinTheta, vzvz * (1 - cosTheta) + cosTheta, 0],
        [0, 0, 0, 1]], dtype=np.float32)

def rotate_x(theta: float): return rotate_angle_axis(theta, v=[1, 0, 0])

def rotate_y(theta: float): return rotate_angle_axis(theta, v=[0, 1, 0])

def rotate_z(theta: float): return rotate_angle_axis(theta, v=[0, 0, 1])

def read_instance_binary_file(fname: str):
    import struct
    with open(fname, 'rb') as f:
        n = struct.unpack('<I', f.read(4))[0]
        instances = []
        for i in range(n):
            row0 = [*struct.unpack('<ffff', f.read(16))]
            row1 = [*struct.unpack('<ffff', f.read(16))]
            row2 = [*struct.unpack('<ffff', f.read(16))]
            instances.append(np.matrix([row0, row1, row2, [0, 0, 0, 1]]))
        return instances

def write_instance_binary_file(fname: str, instances: [np.matrix]):
    import struct
    with open(fname, 'wb') as f:
        f.write(struct.pack('<I', len(instances)))
        for instance in instances:
            f.write(struct.pack('<ffffffffffff', 
                instance[0,0], instance[0,1], instance[0,2], instance[0,3],
                instance[1,0], instance[1,1], instance[1,2], instance[1,3],
                instance[2,0], instance[2,1], instance[2,2], instance[2,3]))

instances = []
instances.append(
  np.matrix([[+2, 0, 0, -5], [0, +2, 0, +5], [0, 0, +2, 0], [0, 0, 0, 1]]))
instances.append(
  np.matrix([[-2, 0, 0, +5], [0, -2, 0, +5], [0, 0, -2, 2], [0, 0, 0, 1]]))
instances.append(
  np.matrix([[+2, 0, 0, -5], [0, +2, 0, -5], [0, 0, +2, 0], [0, 0, 0, 1]]))
instances.append(
  np.matrix([[-2, 0, 0, +5], [0, -2, 0, -5], [0, 0, -2, 2], [0, 0, 0, 1]]))

write_instance_binary_file("test.instances", instances)
